# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from json import (
    dumps as dump_json,
    load as load_json,
    loads as parse_json,
)
from os import (
    remove,
    rename,
)
from os.path import (
    dirname,
    exists,
    islink,
)
from pathlib import Path
from typing import (
    Callable,
    Dict,
    IO,
    List,
    Optional,
    Tuple,
    Union,
)
from uuid import uuid4
from numpy import (
    ndarray,
)
from deareis.data import DataSet
from deareis.data.fitting import FitResult
from deareis.data.drt import DRTResult
from deareis.data.kramers_kronig import TestResult
from deareis.data.simulation import SimulationResult
from deareis.data.plotting import (
    PlotSettings,
    PlotSeries,
)
from deareis.enums import PlotType


VERSION: int = 4


def _parse_v4(state: dict) -> dict:
    # TODO: Update implementation when VERSION is incremented
    return state


def _parse_v3(state: dict) -> dict:
    state["drts"] = {_["uuid"]: [] for _ in state["data_sets"]}
    return _parse_v4(state)


def _parse_v2(state: dict) -> dict:
    state["data_sets"] = state["datasets"]
    key: str
    for key in [
        "active_data_uuid",
        "active_fit_uuid",
        "active_plot_uuid",
        "active_simulation_data_uuid",
        "active_simulation_uuid",
        "active_test_uuid",
        "datasets",
        "latest_fit_circuit",
        "latest_simulation_circuit",
    ]:
        if key in state:
            del state[key]
    if not state["plots"]:
        state["plots"].append(
            PlotSettings(
                "Plot",
                PlotType.NYQUIST,
                [],
                {},
                {},
                {},
                {},
                {},
                uuid4().hex,
            ).to_dict(session=False)
        )
    return _parse_v3(state)


def _parse_v1(state: dict) -> dict:
    state["active_plot_uuid"] = ""
    state["plots"] = []
    return _parse_v2(state)


class Project:
    """
    A class representing a collection of notes, data sets, test results, fit results, simulation results, and complex plots.
    """

    def __init__(self, *args, **kwargs):
        self._path: str = ""
        self.update(*args, **kwargs)

    def __repr(self) -> str:
        return f"Project ({self.get_label()}, {hex(id(self))})"

    def update(self, *args, **kwargs):
        """
        Used when restoring project states.
        """
        if not hasattr(self, "uuid"):
            self.uuid: str = kwargs.get("uuid", uuid4().hex)
        self._data_sets: List[DataSet] = list(
            map(DataSet.from_dict, kwargs.get("data_sets", []))
        )
        self._drts: Dict[str, List[DRTResult]] = {
            k: list(map(DRTResult.from_dict, v))
            for k, v in kwargs.get("drts", {}).items()
        }
        self._fits: Dict[str, List[FitResult]] = {
            k: list(map(FitResult.from_dict, v))
            for k, v in kwargs.get("fits", {}).items()
        }
        self._label: str = kwargs.get("label", "Project")
        self._notes: str = kwargs.get("notes", "")
        path: str = kwargs.get("path", "").strip()
        if path != "":
            self.set_path(path)
        self._plots: List[PlotSettings] = list(
            map(PlotSettings.from_dict, kwargs.get("plots", []))
        )
        if len(self._plots) == 0:
            self._plots.append(
                PlotSettings(
                    "Plot",
                    PlotType.NYQUIST,
                    [],
                    {},
                    {},
                    {},
                    {},
                    {},
                    uuid4().hex,
                )
            )
        self._simulations: List[SimulationResult] = list(
            map(SimulationResult.from_dict, kwargs.get("simulations", []))
        )
        self._tests: Dict[str, List[TestResult]] = {
            k: list(map(TestResult.from_dict, v))
            for k, v in kwargs.get("tests", {}).items()
        }

    @staticmethod
    def parse(state: dict) -> dict:
        """
        Used when deserializing project files.
        """
        assert type(state) is dict, type(state)
        if "version" in state:
            version: int = state["version"]
            assert type(version) is int, version
            assert version >= 1 and version <= VERSION, (
                version,
                VERSION,
            )
            del state["version"]
            parsers: Dict[int, Callable] = {
                1: _parse_v1,
                2: _parse_v2,
                3: _parse_v3,
                4: _parse_v4,
            }
            assert version in parsers, (
                version,
                parsers,
            )
            state = parsers[version](state)
            assert type(state["uuid"]) is str
        # Basic validation
        assert type(state["data_sets"]) is list
        assert type(state["fits"]) is dict
        assert type(state["drts"]) is dict
        assert type(state["label"]) is str
        assert type(state["notes"]) is str
        assert type(state["plots"]) is list
        assert type(state["simulations"]) is list
        assert type(state["tests"]) is dict
        return state

    @classmethod
    def from_dict(Class, state: dict) -> "Project":
        """
        Create an instance from a dictionary.

        Parameters
        ----------
        state: dict
            A dictionary-based representation of a project state.
        """
        return Class(**Class.parse(state))

    @classmethod
    def from_file(Class, path: str) -> "Project":
        """
        Create an instance by parsing a file containing a Project that has been serialized using JSON.

        Parameters
        ----------
        path: str
            The path to a file containing a serialized project state.
        """
        assert type(path) is str and exists(path)
        fp: IO
        with open(path, "r") as fp:
            state: dict = load_json(fp)
        state["path"] = path
        return Class.from_dict(state)

    @classmethod
    def from_json(Class, json: str) -> "Project":
        """
        Create an instance by parsing a JSON string.

        Parameters
        ----------
        json: str
            A JSON representation of a project state.
        """
        assert type(json) is str
        return Class.from_dict(parse_json(json))

    @classmethod
    def merge(Class, projects: List["Project"]) -> "Project":
        """
        Create an instance by merging multiple Project instances.
        All UUIDs are replaced to avoid collisions.
        The labels of some objects are also replaced to avoid collisions.
        """
        assert type(projects) is list and all(
            map(lambda _: type(_) is Class, projects)
        ), projects

        def extract_uuids(dictionary: dict) -> List[str]:
            assert type(dictionary) is dict, dictionary
            uuids: List[str] = []
            for k, v in dictionary.items():
                if type(v) is dict:
                    uuids.extend(extract_uuids(v))
                elif type(v) is list:
                    for v in filter(lambda _: type(_) is dict, v):
                        uuids.extend(extract_uuids(v))
                elif k == "uuid":
                    assert type(v) is str, v
                    uuids.append(v)
            # print(uuids)
            return list(set(uuids))

        def replace_uuids(dictionary: dict) -> dict:
            assert type(dictionary) is dict, dictionary
            uuids: List[str] = extract_uuids(dictionary)
            json: str = dump_json(dictionary)
            uuid: str
            for uuid in uuids:
                json = json.replace(uuid, uuid4().hex)
            return parse_json(json)

        # Merge the various items into one large dictionary.
        # Replace the old UUIDs with new ones to reduce the chance of collisions
        # (e.g. in case one or more of the projects were initially identical copies
        # that were modified in different ways).
        state: dict = Class().to_dict(session=True)
        project: "Project"
        for project in projects:
            other: dict = replace_uuids(project.to_dict(session=True))
            state["data_sets"].extend(other["data_sets"])
            state["fits"].update(other["fits"])
            state["drts"].update(other["drts"])
            state["notes"] = state["notes"] + "\n\n" + other["notes"]
            state["plots"].extend(other["plots"])
            state["simulations"].extend(other["simulations"])
            state["tests"].update(other["tests"])
            state["label"] = other["label"]
        state["notes"] = state["notes"].strip()
        # Check for UUID collisions.
        uuids: List[str] = []
        uuids.extend(list(map(lambda _: _["uuid"], state["data_sets"])))
        for fits in state["fits"].values():
            uuids.extend(list(map(lambda _: _["uuid"], fits)))
        for drts in state["drts"].values():
            uuids.extend(list(map(lambda _: _["uuid"], drts)))
        uuids.extend(list(map(lambda _: _["uuid"], state["plots"])))
        uuids.extend(list(map(lambda _: _["uuid"], state["simulations"])))
        for tests in state["tests"].values():
            uuids.extend(list(map(lambda _: _["uuid"], tests)))
        assert all(map(lambda _: type(_) is str, uuids)), uuids
        assert len(uuids) == len(set(uuids)), "Encountered UUID collision!"
        if len(projects) > 1:
            # Make sure that labels are unique.
            # - Data sets
            labels: List[str] = []
            i: int
            label: str
            data: dict
            for data in state["data_sets"]:
                label = data["label"]
                i = 1
                while label in labels:
                    i += 1
                    label = f"{data['label']} ({i})"
                labels.append(label)
                data["label"] = label
            # - Plot settings
            labels = []
            for plot in state["plots"]:
                label = plot["plot_label"]
                i = 1
                while label in labels:
                    i += 1
                    label = f"{plot['plot_label']} ({i})"
                labels.append(label)
                plot["plot_label"] = label
            # Change the project label
            state["label"] = "Merged project"
        return Class.from_dict(state)

    def to_dict(self, session: bool) -> dict:
        """
        Return a dictionary containing the project state.
        The dictionary can be used to recreate a project or to restore a project state.

        Parameters
        ----------
        session: bool
            If true, then data minimization is not performed.
        """
        return {
            "data_sets": list(
                map(lambda _: _.to_dict(session=session), self._data_sets)
            ),
            "fits": {
                k: list(map(lambda _: _.to_dict(session=session), v))
                for k, v in self._fits.items()
            },
            "drts": {
                k: list(map(lambda _: _.to_dict(), v)) for k, v in self._drts.items()
            },
            "label": self._label,
            "notes": self._notes,
            "plots": list(map(lambda _: _.to_dict(session=session), self._plots)),
            "simulations": list(map(lambda _: _.to_dict(), self._simulations)),
            "tests": {
                k: list(map(lambda _: _.to_dict(session=session), v))
                for k, v in self._tests.items()
            },
            "uuid": self.uuid,
            "version": VERSION,
        }

    def get_label(self) -> str:
        """
        Get the project's label.
        """
        return self._label

    def set_label(self, label: str):
        """
        Set the project's label.

        Parameters
        ----------
        label: str
            The new label.
        """
        assert type(label) is str
        assert label.strip() != ""
        self._label = label

    def set_path(self, path: str):
        """
        Set the path to use when calling the `save` method without arguments.

        Parameters
        ----------
        path: str
            The path where the project's state should be saved.
        """
        assert type(path) is str, path
        self._path = path

    def get_path(self) -> str:
        """
        Get the project's currrent path.
        An empty string signifies that no path has been set previously.
        """
        return self._path

    def get_notes(self) -> str:
        """
        Get the project's notes.
        """
        return self._notes

    def set_notes(self, notes: str):
        """
        Set the project's notes.

        Parameters
        ----------
        notes: str
            The project notes.
        """
        assert type(notes) is str, notes
        self._notes = notes

    def save(self, path: Optional[str] = None):
        """
        Serialize the project as a file containing a JSON string.

        Parameters
        ----------
        path: Optional[str] = None
            The path to write the project state to.
            If this is None, then the most recently defined path is used.
        """
        assert type(path) is str or path is None, path
        if path is None:
            path = self.get_path()
        else:
            self.set_path(path)
        assert path != ""
        assert exists(dirname(path)), path
        suffix: str = f".bak-{uuid4().hex}"
        tmp_path: str = path + suffix
        if exists(path):
            if islink(path):
                tmp_path = str(Path(path).resolve()) + suffix
            rename(path, tmp_path)
        dictionary: dict = self.to_dict(session=False)
        fp: IO
        with open(path, "w") as fp:
            fp.write(dump_json(dictionary, sort_keys=True, indent=1))
        if exists(tmp_path):
            remove(tmp_path)

    def get_data_sets(self) -> List[DataSet]:
        """
        Get the project's data sets.
        """
        return self._data_sets

    def add_data_set(self, data: DataSet):
        """
        Add a data set to the project.

        Parameters
        ----------
        data: DataSet
            The data set to add.
        """
        assert type(data) is DataSet, data
        assert data.uuid not in list(map(lambda _: _.uuid, self._data_sets))
        label: str = data.get_label()
        existing_labels: List[str] = list(map(lambda _: _.get_label(), self._data_sets))
        if label in existing_labels:
            i: int = 1
            while label in existing_labels:
                i += 1
                label = f"{data.get_label()} ({i})"
            data.set_label(label)
        self._data_sets.append(data)
        self._fits[data.uuid] = []
        self._drts[data.uuid] = []
        self._tests[data.uuid] = []
        self._data_sets.sort(key=lambda _: _.get_label())

    def edit_data_set_label(self, data: DataSet, label: str):
        """
        Edit the label of a data set in the project.
        Ensures that each data set has a unique label.

        Parameters
        ----------
        data: DataSet
            The data set to rename.

        label: str
            The new label.
        """
        assert type(data) is DataSet, data
        assert data.uuid in list(map(lambda _: _.uuid, self._data_sets)), data
        label = label.strip()
        if label == data.get_label():
            return
        assert label != "", "The label of a data set cannot be an empty string!"
        assert label not in list(
            map(lambda _: _.get_label(), self.get_data_sets())
        ), f"Another data set already has the label '{label}'!"
        data.set_label(label)
        self._data_sets.sort(key=lambda _: _.get_label())

    def edit_data_set_path(self, data: DataSet, path: str):
        """
        Edit the path of a data set in the project.

        Parameters
        ----------
        data: DataSet
            The data set to edit.

        path: str
            The new path.
        """
        assert type(data) is DataSet, data
        assert data.uuid in list(map(lambda _: _.uuid, self._data_sets)), data
        assert type(path) is str, path
        data.set_path(path)

    def delete_data_set(self, data: DataSet):
        """
        Delete a data set (and its associated test and fit results) from the project.

        Parameters
        ----------
        data: DataSet
            The data set to remove.
        """
        assert type(data) is DataSet, data
        assert data.uuid in list(map(lambda _: _.uuid, self._data_sets)), data
        self._data_sets.remove(data)
        del self._fits[data.uuid]
        del self._drts[data.uuid]
        del self._tests[data.uuid]
        list(map(lambda _: _.remove_series(data.uuid), self._plots))

    def replace_data_set(self, old: DataSet, new: DataSet):
        """
        Replace a data set in the project with another one.

        Parameters
        ----------
        old: DataSet
            The data set to be replaced.

        new: DataSet
            The replacement data set.
        """
        assert type(old) is DataSet, old
        assert type(new) is DataSet, new
        assert old.uuid in list(map(lambda _: _.uuid, self._data_sets))
        assert old.uuid == new.uuid, (
            old.uuid,
            new.uuid,
        )
        self._data_sets.remove(old)
        self._data_sets.append(new)
        self._data_sets.sort(key=lambda _: _.get_label())

    def get_all_tests(self) -> Dict[str, List[TestResult]]:
        """
        Get a mapping of data set UUIDs to the corresponding Kramers-Kronig test results of those data sets.
        """
        return self._tests

    def get_tests(self, data: DataSet) -> List[TestResult]:
        """
        Get the Kramers-Kronig test results associated with a specific data set.

        Parameters
        ----------
        data: DataSet
            The data set whose tests to get.
        """
        assert type(data) is DataSet, data
        assert data.uuid in list(map(lambda _: _.uuid, self._data_sets)), data
        return self._tests[data.uuid]

    def add_test(self, data: DataSet, test: TestResult):
        """
        Add the provided Kramers-Kronig test result to the provided data set's list of Kramers-Kronig test results.

        Parameters
        ----------
        data: DataSet
            The data set that was tested.

        test: TestResult
            The result of the test.
        """
        assert type(data) is DataSet, data
        assert data.uuid in list(map(lambda _: _.uuid, self._data_sets)), data
        assert type(test) is TestResult, test
        assert test.uuid not in list(map(lambda _: _.uuid, self._tests[data.uuid]))
        self._tests[data.uuid].insert(0, test)

    def delete_test(self, data: DataSet, test: TestResult):
        """
        Delete the provided Kramers-Kronig test result from the provided data set's list of Kramers-Kronig test results.

        Parameters
        ----------
        data: DataSet
            The data set associated with the test result.

        test: TestResult
            The test result to delete.
        """
        assert type(data) is DataSet, data
        assert data.uuid in list(map(lambda _: _.uuid, self._data_sets)), data
        assert type(test) is TestResult, test
        assert test in self._tests[data.uuid], test
        self._tests[data.uuid].remove(test)
        list(map(lambda _: _.remove_series(test.uuid), self._plots))

    def get_all_drts(self) -> Dict[str, List[DRTResult]]:
        """
        Get a mapping of data set UUIDs to the corresponding DRT analysis results of those data sets.
        """
        return self._drts

    def get_drts(self, data: DataSet) -> List[DRTResult]:
        """
        Get the DRT analysis results associated with a specific data set.

        Parameters
        ----------
        data: DataSet
            The data set whose analyses to get.
        """
        assert type(data) is DataSet, data
        assert data.uuid in list(map(lambda _: _.uuid, self._data_sets)), data
        return self._drts[data.uuid]

    def add_drt(self, data: DataSet, drt: DRTResult):
        """
        Add the provided DRT analysis result to the provided data set's list of DRT analysis results.

        Parameters
        ----------
        data: DataSet
            The data set that was analyzed.

        drt: DRTResult
            The result of the analysis.
        """
        assert type(data) is DataSet, data
        assert data.uuid in list(map(lambda _: _.uuid, self._data_sets)), data
        assert type(drt) is DRTResult, drt
        assert drt.uuid not in list(map(lambda _: _.uuid, self._drts[data.uuid]))
        self._drts[data.uuid].insert(0, drt)

    def delete_drt(self, data: DataSet, drt: DRTResult):
        """
        Delete the provided DRT analysis result from the provided data set's list of DRT analysis results.

        Parameters
        ----------
        data: DataSet
            The data set associated with the analysis result.

        drt: DRTResult
            The analysis result to delete.
        """
        assert type(data) is DataSet, data
        assert data.uuid in list(map(lambda _: _.uuid, self._data_sets)), data
        assert type(drt) is DRTResult, drt
        assert drt in self._drts[data.uuid], drt
        self._drts[data.uuid].remove(drt)
        list(map(lambda _: _.remove_series(drt.uuid), self._plots))

    def get_all_fits(self) -> Dict[str, List[FitResult]]:
        """
        Get a mapping of data set UUIDs to the corresponding list of fit results of those data sets.
        """
        return self._fits

    def get_fits(self, data: DataSet) -> List[FitResult]:
        """
        Get fit results associated with a specific data set.

        Parameters
        ----------
        data: DataSet
            The data set whose fits to get.
        """
        assert type(data) is DataSet, data
        assert data.uuid in list(map(lambda _: _.uuid, self._data_sets)), data
        return self._fits[data.uuid]

    def add_fit(self, data: DataSet, fit: FitResult):
        """
        Add the provided fit result to the provided data set.

        Parameters
        ----------
        data: DataSet
            The data set that a circuit was fit to.

        fit: FitResult
            The result of the circuit fit.
        """
        assert type(data) is DataSet, data
        assert data.uuid in list(map(lambda _: _.uuid, self._data_sets)), data
        assert type(fit) is FitResult, fit
        assert fit.uuid not in list(map(lambda _: _.uuid, self._fits[data.uuid]))
        self._fits[data.uuid].insert(0, fit)

    def delete_fit(self, data: DataSet, fit: FitResult):
        """
        Delete the provided fit result from the provided data set's list of fit results.

        Parameters
        ----------
        data: DataSet
            The data set associated with the fit result.

        fit: FitResult
            The fit result to delete.
        """
        assert type(data) is DataSet, data
        assert data.uuid in list(map(lambda _: _.uuid, self._data_sets)), data
        assert type(fit) is FitResult, fit
        assert fit in self._fits[data.uuid], fit
        self._fits[data.uuid].remove(fit)
        list(map(lambda _: _.remove_series(fit.uuid), self._plots))

    def get_simulations(self) -> List[SimulationResult]:
        """
        Get all of the simulation results.
        """
        return self._simulations

    def add_simulation(self, simulation: SimulationResult):
        """
        Add the provided simulation result to the list of simulation results.

        Parameters
        ----------
        simulation: SimulationResult
            The result of the simulation.
        """
        assert type(simulation) is SimulationResult, simulation
        assert simulation.uuid not in list(map(lambda _: _.uuid, self._simulations))
        self._simulations.insert(0, simulation)

    def delete_simulation(self, simulation: SimulationResult):
        """
        Remove the provided simulation result from the list of simulation results.

        Parameters
        ----------
        simulation: SimulationResult
            The simulation result to delete.
        """
        assert type(simulation) is SimulationResult, simulation
        assert simulation in self._simulations
        self._simulations.remove(simulation)
        list(map(lambda _: _.remove_series(simulation.uuid), self._plots))

    def get_plots(self) -> List[PlotSettings]:
        """
        Get all of the plots.
        """
        return self._plots

    def add_plot(self, plot: PlotSettings):
        """
        Add the provided plot to the list of plots.

        Parameters
        ----------
        plot: PlotSettings
            The settings for the plot.
        """
        assert type(plot) is PlotSettings, plot
        assert plot.uuid not in list(map(lambda _: _.uuid, self._plots))
        self._plots.append(plot)
        self._plots.sort(key=lambda _: _.get_label())

    def edit_plot_label(self, plot: PlotSettings, label: str):
        """
        Edit the label of a plot in the project.
        Ensures that each plot has a unique label.

        Parameters
        ----------
        plot: PlotSettings
            The plot settings to edit.

        label: str
            The new label.
        """
        assert type(plot) is PlotSettings, plot
        assert plot in self._plots, plot
        label = label.strip()
        if label == plot.get_label():
            return
        assert label != "", "The label of a plot cannot be an empty string!"
        assert label not in list(
            map(lambda _: _.get_label(), self.get_plots())
        ), f"Another plot already has the label '{label}'!"
        plot.set_label(label)
        self._plots.sort(key=lambda _: _.get_label())

    def delete_plot(self, plot: PlotSettings):
        """
        Delete the provided plot from the list of plots.

        Parameters
        ----------
        plot: PlotSettings
            The plot settings to delete.
        """
        assert type(plot) is PlotSettings, plot
        assert plot in self._plots, plot
        self._plots.remove(plot)

    def get_plot_series(
        self,
        plot: PlotSettings,
    ) -> List[PlotSeries]:
        """
        Get PlotSeries instances of each of the plotted items/series in a specific plot.

        Parameters
        ----------
        plot: PlotSettings
            The plot whose items/series to get.
        """
        assert type(plot) is PlotSettings, plot
        data_sets: List[DataSet] = self.get_data_sets()
        tests: Dict[str, List[TestResult]] = self.get_all_tests()
        drts: Dict[str, List[DRTResult]] = self.get_all_drts()
        fits: Dict[str, List[FitResult]] = self.get_all_fits()
        simulations: List[SimulationResult] = self.get_simulations()
        results: List[PlotSeries] = []
        uuid: str
        for uuid in plot.series_order:
            series: Optional[
                Union[DataSet, TestResult, DRTResult, FitResult, SimulationResult]
            ]
            series = plot.find_series(uuid, data_sets, tests, drts, fits, simulations)
            if series is None:
                continue
            label: str = plot.get_series_label(uuid) or series.get_label()
            scatter_data: Tuple[ndarray, ndarray]
            line_data: Tuple[ndarray, ndarray]
            fill_data: Tuple[ndarray, ndarray, ndarray]
            color: Tuple[float, float, float, float] = tuple(
                map(
                    lambda _: _ / 255.0,
                    plot.get_series_color(uuid) or [0.0, 0.0, 0.0, 255.0],
                )
            )
            marker: int = plot.get_series_marker(uuid)
            line: bool = plot.get_series_line(uuid)
            results.append(
                PlotSeries(
                    series,
                    label,
                    color,
                    marker,
                    line,
                    label.strip() != "",
                )
            )
        return results



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from argparse import (
    ArgumentParser,
    Namespace,
    RawTextHelpFormatter,
)
from os.path import abspath
from deareis.version import PACKAGE_VERSION


def parse() -> Namespace:
    parser: ArgumentParser = ArgumentParser(
        allow_abbrev=False,
        formatter_class=RawTextHelpFormatter,
        description=f"""
DearEIS ({PACKAGE_VERSION})
A GUI program for analyzing, simulating, and visualizing impedance spectra.

Copyright (C) 2022 DearEIS developers

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY
APPLICABLE LAW.  EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT
HOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM "AS IS" WITHOUT WARRANTY
OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
PURPOSE.  THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM
IS WITH YOU.  SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF
ALL NECESSARY SERVICING, REPAIR OR CORRECTION.
""".strip(),
    )
    parser.add_argument(
        "-d",
        "--data-files",
        metavar="path",
        dest="data_files",
        nargs="*",
        default=[],
        help="Create a new project and load the specified data files as data sets in that project.",
    )
    parser.add_argument(
        "project_files",
        nargs="*",
        metavar="project",
        default=[],
        help="Load the specified project files.",
    )
    args: Namespace = parser.parse_args()
    if len(args.data_files) > 0:
        args.data_files = list(map(abspath, args.data_files))
    if len(args.project_files) > 0:
        args.project_files = list(map(abspath, args.project_files))
    return args



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

import dearpygui.dearpygui as dpg
from deareis.utility import calculate_window_position_dimensions
from deareis.signals import Signal
import deareis.signals as signals


class ErrorMessage:
    def __init__(self):
        self.window: int = dpg.generate_uuid()
        with dpg.window(
            label="ERROR",
            modal=True,
            no_resize=True,
            show=False,
            width=720,
            height=540,
            on_close=self.hide,
            tag=self.window,
        ):
            self.message_text: int = dpg.generate_uuid()
            dpg.add_text(tag=self.message_text)
            with dpg.child_window(width=-1, height=-24, horizontal_scrollbar=True):
                self.traceback_text: int = dpg.generate_uuid()
                dpg.add_text(tag=self.traceback_text)
            dpg.add_button(
                label="Copy to clipboard",
                callback=lambda: dpg.set_clipboard_text(
                    dpg.get_value(self.traceback_text)
                ),
            )

    def is_visible(self) -> bool:
        return dpg.is_item_visible(self.window)

    def show(self, traceback: str, message: str = ""):
        assert type(traceback) is str
        assert type(message) is str
        print(f"\n{traceback}")
        dpg.split_frame(delay=33)
        if message.strip() == "":
            message = """
An exception has been raised! The very end of the traceback (see below) may offer a hint on how to solve the cause for the error that just occurred. If it doesn't, then please copy the traceback to your clipboard and include it in a bug report. Bug reports can be submitted at github.com/vyrjana/DearEIS/issues.
            """.strip()
        if not self.is_visible():
            dpg.show_item(self.window)
            dpg.split_frame()
            self.key_handler: int = dpg.generate_uuid()
            with dpg.handler_registry(tag=self.key_handler):
                dpg.add_key_release_handler(
                    key=dpg.mvKey_Escape,
                    callback=self.hide,
                )
        else:
            traceback = f"{dpg.get_value(self.traceback_text)}\n\n{traceback}"
        width: int
        width, _ = dpg.get_item_rect_size(self.window)
        dpg.configure_item(self.message_text, wrap=width)
        if message != dpg.get_value(self.message_text):
            dpg.set_value(self.message_text, message)
        dpg.set_value(self.traceback_text, traceback)

    def hide(self):
        if dpg.does_item_exist(self.key_handler):
            dpg.delete_item(self.key_handler)
        if dpg.is_item_visible(self.window):
            dpg.hide_item(self.window)
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)

    def resize(self, width: int, height: int):
        assert type(width) is int
        assert type(height) is int
        x: int
        y: int
        x, y, width, height = calculate_window_position_dimensions()
        dpg.configure_item(
            self.window,
            pos=(
                x,
                y,
            ),
            width=width,
            height=height,
        )
        dpg.split_frame()
        width, _ = dpg.get_item_rect_size(self.window)
        dpg.configure_item(self.message_text, wrap=width)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from os.path import dirname
from traceback import format_exc
from typing import (
    Dict,
    List,
    Optional,
)
from pyimpspec.data.formats import UnsupportedFileFormat
from pyimpspec.data import get_parsers
import pyimpspec
from deareis.data import (
    DRTResult,
    DataSet,
    FitResult,
    Project,
    SimulationResult,
    TestResult,
)
from deareis.enums import Context
from deareis.gui import ProjectTab
from deareis.gui.data_sets.average_data_sets import AverageDataSets
from deareis.gui.data_sets.copy_mask import CopyMask
from deareis.gui.data_sets.subtract_impedance import SubtractImpedance
from deareis.gui.data_sets.toggle_data_points import ToggleDataPoints
from deareis.gui.file_dialog import FileDialog
from deareis.signals import Signal
from deareis.state import STATE
import deareis.signals as signals


def select_data_set(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    is_busy_message_visible: bool = STATE.is_busy_message_visible()
    if not is_busy_message_visible:
        signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Loading data set")
    data: Optional[DataSet] = kwargs.get("data")
    project_tab.select_data_set(data)
    project_tab.populate_tests(project, data)
    project_tab.populate_drts(project, data)
    project_tab.populate_fits(project, data)
    project_tab.populate_simulations(project)
    if not is_busy_message_visible:
        signals.emit(Signal.HIDE_BUSY_MESSAGE)


def load_simulation_as_data_set(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    simulation: SimulationResult = kwargs.get("simulation")
    if simulation is None:
        return
    signals.emit(
        Signal.SHOW_BUSY_MESSAGE,
        message="Loading simulation result as data set",
    )
    data: DataSet = DataSet(
        simulation.get_frequency(),
        simulation.get_impedance(),
        mask={},
        label=simulation.get_label(),
    )
    project.add_data_set(data)
    project_tab.populate_data_sets(project)
    signals.emit(Signal.SELECT_DATA_SET, data=data)
    signals.emit(
        Signal.SELECT_PLOT_SETTINGS,
        settings=project_tab.get_active_plot(),
    )
    signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)
    signals.emit(Signal.HIDE_BUSY_MESSAGE)


def load_data_set_files(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    paths: List[str] = kwargs.get("paths")
    if not paths:
        return
    existing_labels: List[str] = list(
        map(lambda _: _.get_label(), project.get_data_sets())
    )
    parsing_errors: Dict[str, str] = {}
    loaded_data: bool = False
    num_paths: int = len(paths)
    n: int
    path: str
    for n, path in enumerate(paths):
        signals.emit(
            Signal.SHOW_BUSY_MESSAGE,
            message=f"Loading data: {n + 1}/{num_paths}",
            progress=n / num_paths,
        )
        try:
            data_sets: List[pyimpspec.DataSet] = pyimpspec.parse_data(path)
        except UnsupportedFileFormat:
            parsing_errors[path] = "Unsupported file type!\n"
            continue
        except Exception:
            parsing_errors[path] = format_exc()
            continue
        data: DataSet
        for data in map(lambda _: DataSet.from_dict(_.to_dict()), data_sets):
            label: str = data.get_label().strip()
            if label == "":
                label = "Data set"
            i: int = 0
            while label in existing_labels:
                i += 1
                label = f"{data.get_label()} ({i})"
            existing_labels.append(label)
            data.set_label(label)
            project.add_data_set(data)
        loaded_data = True
    signals.emit(Signal.HIDE_BUSY_MESSAGE)
    STATE.latest_data_set_directory = dirname(path)
    if loaded_data:
        project_tab.populate_data_sets(project)
        signals.emit(Signal.SELECT_DATA_SET, data=data)
        signals.emit(
            Signal.SELECT_PLOT_SETTINGS,
            settings=project_tab.get_active_plot(),
        )
        signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)
    if parsing_errors:
        total_traceback: str = ""
        traceback: str
        for path, traceback in parsing_errors.items():
            total_traceback += f"{traceback}\nThe exception above was encountered while parsing '{path}'.\n\n"
        signals.emit(
            Signal.SHOW_ERROR_MESSAGE,
            traceback=total_traceback.strip(),
            message="""
Encountered error(s) while parsing data file(s). The file(s) might be malformed, corrupted, or not supported by this version of DearEIS.
            """.strip(),
        )


def select_data_set_files(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    FileDialog(
        cwd=STATE.latest_data_set_directory,
        label="Select data file(s)",
        callback=lambda *a, **k: signals.emit(Signal.LOAD_DATA_SET_FILES, *a, **k),
        extensions=[".*"] + list(get_parsers().keys()),
    )


def rename_data_set(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    label: Optional[str] = kwargs.get("label")
    data: Optional[DataSet] = kwargs.get("data")
    if label is None or data is None:
        return
    if label.strip() == data.get_label():
        return
    project.edit_data_set_label(data, label)
    project_tab.populate_data_sets(project)
    signals.emit(Signal.SELECT_PLOT_SETTINGS, settings=project_tab.get_active_plot())
    signals.emit(Signal.SELECT_DATA_SET, data=data)
    signals.emit(
        Signal.SELECT_SIMULATION_RESULT,
        simulation=project_tab.get_active_simulation(),
        data=project_tab.get_active_data_set(context=Context.SIMULATION_TAB),
    )
    signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)


def modify_data_set_path(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    path: Optional[str] = kwargs.get("path")
    data: Optional[DataSet] = kwargs.get("data")
    if path is None or data is None:
        return
    project.edit_data_set_path(data, path.strip())
    signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)


def delete_data_set(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    data: Optional[DataSet] = kwargs.get("data")
    if data is None:
        return
    signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Deleting data set")
    project.delete_data_set(data)
    project_tab.populate_data_sets(project)
    signals.emit(Signal.SELECT_PLOT_SETTINGS, settings=project_tab.get_active_plot())
    lookup: Dict[str, DataSet] = {_.get_label(): _ for _ in project.get_data_sets()}
    labels: List[str] = list(lookup.keys())
    signals.emit(
        Signal.SELECT_DATA_SET,
        data=lookup[labels[0]] if labels else None,
    )
    data = project_tab.get_active_data_set(context=Context.SIMULATION_TAB)
    if data is not None and data.get_label() not in labels:
        data = None
    signals.emit(
        Signal.SELECT_SIMULATION_RESULT,
        simulation=project_tab.get_active_simulation(),
        data=data,
    )
    signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)
    signals.emit(Signal.HIDE_BUSY_MESSAGE)


def toggle_data_point(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    state: Optional[bool] = kwargs.get("state")
    index: Optional[int] = kwargs.get("index")
    data: Optional[DataSet] = kwargs.get("data")
    if state is None or index is None or data is None:
        return
    signals.emit(
        Signal.SHOW_BUSY_MESSAGE,
        message="Toggling data point",
    )
    mask: Dict[int, bool] = data.get_mask()
    mask[index] = state
    data.set_mask(mask)
    signals.emit(Signal.SELECT_PLOT_SETTINGS, settings=project_tab.get_active_plot())
    signals.emit(Signal.SELECT_DATA_SET, data=data)
    signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)
    signals.emit(Signal.HIDE_BUSY_MESSAGE)


def apply_data_set_mask(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    mask: Optional[Dict[int, bool]] = kwargs.get("mask")
    data: Optional[DataSet] = kwargs.get("data")
    if mask is None or data is None:
        return
    data.set_mask(mask)
    signals.emit(Signal.SELECT_PLOT_SETTINGS, settings=project_tab.get_active_plot())
    signals.emit(Signal.SELECT_DATA_SET, data=data)
    test: Optional[TestResult] = kwargs.get("test")
    drt: Optional[DRTResult] = kwargs.get("drt")
    fit: Optional[FitResult] = kwargs.get("fit")
    if test is not None:
        signals.emit(Signal.SELECT_TEST_RESULT, data=data, test=test)
    elif drt is not None:
        signals.emit(Signal.SELECT_DRT_RESULT, data=data, drt=drt)
    elif fit is not None:
        signals.emit(Signal.SELECT_FIT_RESULT, data=data, fit=fit)
    signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)


def select_data_sets_to_average(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return

    def add_data(data: DataSet):
        assert project is not None
        assert project_tab is not None
        project.add_data_set(data)
        project_tab.populate_data_sets(project)
        signals.emit(
            Signal.SELECT_PLOT_SETTINGS, settings=project_tab.get_active_plot()
        )
        signals.emit(Signal.SELECT_DATA_SET, data=data)
        signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)

    data_sets: List[DataSet] = project.get_data_sets()
    if len(data_sets) < 2:
        return

    average_data_sets_window: AverageDataSets = AverageDataSets(
        data_sets=data_sets,
        callback=add_data,
    )
    signals.emit(
        Signal.BLOCK_KEYBINDINGS,
        window=average_data_sets_window.window,
        window_object=average_data_sets_window,
    )


def select_data_points_to_toggle(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    data: Optional[DataSet] = kwargs.get("data")
    if project is None or project_tab is None or data is None:
        return
    toggle_data_points_window: ToggleDataPoints = ToggleDataPoints(
        data=data,
        callback=lambda m: signals.emit(
            Signal.APPLY_DATA_SET_MASK,
            mask=m,
            data=data,
        ),
    )
    signals.emit(
        Signal.BLOCK_KEYBINDINGS,
        window=toggle_data_points_window.window,
        window_object=toggle_data_points_window,
    )


def select_data_set_mask_to_copy(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    data: Optional[DataSet] = kwargs.get("data")
    if project is None or project_tab is None or data is None:
        return
    data_sets: List[DataSet] = project.get_data_sets()
    if len(data_sets) < 2:
        return
    copy_mask_window: CopyMask = CopyMask(
        data=data,
        data_sets=data_sets,
        callback=lambda m: signals.emit(
            Signal.APPLY_DATA_SET_MASK,
            mask=m,
            data=data,
        ),
    )
    signals.emit(
        Signal.BLOCK_KEYBINDINGS,
        window=copy_mask_window.window,
        window_object=copy_mask_window,
    )


def select_impedance_to_subtract(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    data: Optional[DataSet] = kwargs.get("data")
    if project is None or project_tab is None or data is None:
        return

    def replace_data(new: DataSet):
        assert project is not None
        assert project_tab is not None
        assert data is not None
        project.replace_data_set(data, new)
        project_tab.populate_data_sets(project)
        signals.emit(
            Signal.SELECT_PLOT_SETTINGS, settings=project_tab.get_active_plot()
        )
        signals.emit(Signal.SELECT_DATA_SET, data=new)
        signals.emit(
            Signal.SELECT_SIMULATION_RESULT,
            simulation=project_tab.get_active_simulation(),
            data=project_tab.get_active_data_set(context=Context.SIMULATION_TAB),
        )
        signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)

    subtract_impedance_window: SubtractImpedance = SubtractImpedance(
        data=data,
        data_sets=project.get_data_sets(),
        fits=project.get_fits(data),
        callback=replace_data,
    )
    signals.emit(
        Signal.BLOCK_KEYBINDINGS,
        window=subtract_impedance_window.window,
        window_object=subtract_impedance_window,
    )



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import Callable, Dict, List, Optional, Set, Tuple, Union
import dearpygui.dearpygui as dpg
from pyimpspec import Circuit, Connection, Element, Parallel, Series, ParsingError
import pyimpspec
import deareis.themes as themes


class Node:
    def __init__(
        self,
        parent: int,
        node_id: int,
        label: str,
        pos: Tuple[int, int],
        element: Optional[Element] = None,
        input_attribute: bool = True,
        output_attribute: bool = True,
    ):
        assert type(parent) is int
        assert type(node_id) is int
        assert type(label) is str
        assert (
            type(pos) is tuple
            and len(pos) == 2
            and all(map(lambda _: type(_) is int, pos))
        )
        assert isinstance(element, Element) or element is None
        assert type(input_attribute) is bool
        assert type(output_attribute) is bool
        self.tag: int = dpg.generate_uuid()
        self.id: int = node_id
        self.label: str = ""
        self.element: Optional[Element] = element
        self.input_attribute: int = dpg.generate_uuid() if input_attribute else -1
        self.output_attribute: int = dpg.generate_uuid() if output_attribute else -1
        self.input_links: Dict[int, int] = {}
        self.output_links: Dict[int, int] = {}
        with dpg.node(label=label, pos=pos, tag=self.tag, parent=parent):
            if input_attribute:
                dpg.add_node_attribute(
                    attribute_type=dpg.mvNode_Attr_Input, tag=self.input_attribute
                )
            if output_attribute:
                dpg.add_node_attribute(
                    attribute_type=dpg.mvNode_Attr_Output, tag=self.output_attribute
                )
        self.set_label(label)

    def __repr__(self) -> str:
        return self.label.strip()

    def set_label(self, label: str):
        assert type(label) is str
        width: int = 5
        label = label.ljust(width, " ")
        self.label = label
        dpg.configure_item(self.tag, label=label[:width])

    def has_attribute(self, attribute: int) -> bool:
        assert type(attribute) is int
        if self.input_attribute >= 0 and self.input_attribute == attribute:
            return True
        elif self.output_attribute >= 0 and self.output_attribute == attribute:
            return True
        return False

    def get_num_input_links(self) -> int:
        return len(self.input_links)

    def get_num_output_links(self) -> int:
        return len(self.output_links)

    def has_link_to(self, link: int) -> bool:
        assert type(link) is int
        return link in self.output_links.values()

    def has_link_from(self, link: int) -> bool:
        assert type(link) is int
        return link in self.input_links.values()

    def add_link_to(self, node: "Node", link: int):
        assert type(link) is int
        self.output_links[node.id] = link

    def delete_link_to(self, node: "Node") -> int:
        link: int = self.output_links[node.id]
        del self.output_links[node.id]
        return link

    def add_link_from(self, node: "Node", link: int):
        assert type(link) is int
        self.input_links[node.id] = link

    def delete_link_from(self, node: "Node") -> int:
        link: int = self.input_links[node.id]
        del self.input_links[node.id]
        return link

    def set_valid(self):
        dpg.bind_item_theme(self.tag, themes.circuit_editor.valid_node)

    def set_invalid(self, msg: str) -> str:
        dpg.bind_item_theme(self.tag, themes.circuit_editor.invalid_node)
        return msg


class Parser:
    def __init__(self, node_editor: int, node_handler: int = -1):
        assert type(node_editor) is int, node_editor
        assert dpg.does_item_exist(node_editor) and "mvNodeEditor" in dpg.get_item_type(
            node_editor
        )
        assert type(node_handler) is int, node_handler
        assert node_handler < 0 or dpg.does_item_exist(node_handler), node_handler
        self.node_editor: int = node_editor
        dpg.configure_item(
            self.node_editor,
            callback=self.link,
            delink_callback=self.delink,
        )
        self.node_handler: int = node_handler
        self.we_node: Node = None  # type: ignore
        self.cere_node: Node = None  # type: ignore
        self.nodes: List[Node] = []
        self.element_counter: int = -1
        self.dummy_counter: int = -1
        self.x_offset: int = 10
        self.y_offset: int = 10
        self.x_step: int = 90
        self.y_step: int = 70
        self.elements: Dict[str, Element] = {
            _.get_description(): _ for _ in pyimpspec.get_elements().values()
        }

    def circuit_to_nodes(self, circuit: Circuit):
        assert type(circuit) is Circuit, circuit
        input_stack: List[Tuple[str, Union[Element, Connection]]] = circuit.to_stack()
        assert circuit.to_string() == "".join(map(lambda _: _[0], input_stack))
        self.clear_nodes()
        if circuit.to_string() not in ["[]", "()"]:
            self.generate_nodes(input_stack)

    def nodes_to_circuit(self, node_editor: int) -> Circuit:
        assert type(node_editor) is int, node_editor
        circuit: Optional[Circuit]
        msg: str
        stack: List[str]
        circuit, msg, stack = self.generate_circuit()
        assert type(circuit) is Circuit, circuit
        return circuit

    def find_node(
        self,
        tag: int = -1,
        node_id: int = -1,
        attribute: int = -1,
        link_to: int = -1,
        link_from: int = -1,
    ) -> Node:
        assert type(tag) is int, tag
        assert type(node_id) is int, node_id
        assert type(attribute) is int, attribute
        assert type(link_to) is int, link_to
        assert type(link_from) is int, link_from
        func: Callable
        if tag >= 0:
            func = lambda _: _.tag == tag
        elif node_id != -1:
            func = lambda _: _.id == node_id
        elif attribute >= 0:
            func = lambda _: _.has_attribute(attribute)
        elif link_to >= 0:
            func = lambda _: _.has_link_to(link_to)
        elif link_from >= 0:
            func = lambda _: _.has_link_from(link_from)
        else:
            raise Exception("Nothing has been provided to use to find a node!")
        node: Node
        for node in self.nodes:
            if func(node):
                return node
        if func(self.we_node):
            return self.we_node
        elif func(self.cere_node):
            return self.cere_node
        raise Exception("Node does not exist!")

    def link(self, sender: int, attributes: Tuple[int, int]):
        assert type(sender) is int
        assert (
            type(attributes) is tuple
            and len(attributes) == 2
            and all(map(lambda _: type(_) is int, attributes))
        )
        link: int = dpg.add_node_link(*attributes, parent=sender)
        src: Node = self.find_node(attribute=attributes[0])
        dst: Node = self.find_node(attribute=attributes[1])
        # print(f"Link: {src.label=}, {dst.label=}, {link=}, {validate=}")
        src.add_link_to(dst, link)
        dst.add_link_from(src, link)

    def delink(self, sender: int, link: int):
        assert type(sender) is int
        assert type(link) is int
        src: Node = self.find_node(link_to=link)
        dst: Node = self.find_node(link_from=link)
        # print(f"Delink: {src.label=}, {dst.label=}, {link=}")
        src.delete_link_to(dst)
        dst.delete_link_from(src)
        dpg.delete_item(link)

    def link_nodes(self, src: Node, dst: Node):
        assert type(src) is Node
        assert type(dst) is Node
        src_attr = dpg.get_item_children(src.tag, slot=1)[
            1 if src != self.we_node else 0
        ]
        dst_attr = dpg.get_item_children(dst.tag, slot=1)[0]
        self.link(self.node_editor, (src_attr, dst_attr))

    def add_node(self, **kwargs) -> Node:
        element: Optional[Element] = kwargs.get("element", None)
        pos: Tuple[int, int] = kwargs.get(
            "pos",
            (
                0,
                1,
            ),
        )
        pos = (
            self.x_offset + self.x_step * pos[0],
            self.y_offset + self.y_step * pos[1],
        )
        self.nodes.append(
            Node(
                parent=self.node_editor,
                node_id=kwargs.get("node_id", len(self.nodes)),
                label=kwargs.get("label", "?"),
                pos=pos,
                element=element,
            )
        )
        if self.node_handler > 0:
            dpg.bind_item_handler_registry(self.nodes[-1].tag, self.node_handler)
        return self.nodes[-1]

    def add_element_node(self, element: Element, **kwargs) -> Node:
        assert isinstance(element, Element)
        kwargs["element"] = element
        if "label" not in kwargs:
            kwargs["label"] = element.get_label()
        node_id: int = kwargs.get("node_id", -1)
        if node_id < 0:
            node_id = self.next_element()
            kwargs["node_id"] = node_id
        element._assign_identifier(node_id)
        node: Node = self.add_node(**kwargs)
        node.set_label(element.get_label())
        return node

    def add_dummy_node(self, **kwargs) -> Node:
        kwargs["label"] = "Dummy"
        node_id: int = kwargs.get("node_id", -1)
        if node_id >= -1:
            kwargs["node_id"] = self.next_dummy()
        return self.add_node(**kwargs)

    def delete_node(self, node: Node):
        assert type(node) is Node
        other: Node
        other_id: int
        link: int
        for other_id, link in node.output_links.copy().items():
            other = self.find_node(node_id=other_id)
            node.delete_link_to(other)
            other.delete_link_from(node)
            dpg.delete_item(link)
        for other_id, link in node.input_links.copy().items():
            other = self.find_node(node_id=other_id)
            node.delete_link_from(other)
            other.delete_link_to(node)
            dpg.delete_item(link)
        if node in self.nodes:
            self.nodes.remove(node)
        dpg.delete_item(node.tag)

    def clear_nodes(self):
        self.reset_element_counter()
        self.reset_dummy_counter()
        dpg.delete_item(self.node_editor, children_only=True)
        self.we_node = Node(
            self.node_editor,
            -999999,
            "WE",
            (
                self.x_offset,
                self.y_offset,
            ),
            input_attribute=False,
        )
        if self.node_handler > 0:
            dpg.bind_item_handler_registry(self.we_node.tag, self.node_handler)
        self.cere_node = Node(
            self.node_editor,
            999999,
            "CE+RE",
            (
                self.x_offset + self.x_step * 2,
                self.y_offset,
            ),
            output_attribute=False,
        )
        if self.node_handler > 0:
            dpg.bind_item_handler_registry(self.cere_node.tag, self.node_handler)
        self.nodes = []

    def reset_element_counter(self):
        self.element_counter = -1

    def next_element(self) -> int:
        self.element_counter += 1
        assert self.element_counter < self.cere_node.id
        return self.element_counter

    def reset_dummy_counter(self):
        self.dummy_counter = -1

    def next_dummy(self) -> int:
        self.dummy_counter -= 1
        assert self.dummy_counter > self.we_node.id
        return self.dummy_counter

    def walk_nodes(
        self,
        node: Node,
        stack: List[str],
        visited_nodes: Set[int],
        pending_nodes: Set[int],
    ):
        assert type(node) is Node
        assert type(stack) is list and all(map(lambda _: type(_) is str, stack))
        assert type(visited_nodes) is set
        assert type(pending_nodes) is set
        if node.id in visited_nodes:
            return
        num_links_out: int = len(node.output_links)
        num_links_in: int
        stack_len: int
        if node == self.we_node:  # WE node (i.e. the starting point)
            assert num_links_out > 0, self.we_node.set_invalid(
                "WE is not connected to anything!"
            )
            assert (
                self.cere_node.id not in node.output_links
            ), self.we_node.set_invalid(
                self.cere_node.set_invalid("WE is shorted to CE+RE!")
            )
            stack.append("[")
            if num_links_out > 1:  # Start of parallel connection
                stack.append("(")
                tmp_pending_nodes = set(pending_nodes)
                for ident, link in node.output_links.items():
                    stack.append("[")
                    stack_len = len(stack)
                    self.walk_nodes(
                        self.find_node(node_id=ident),
                        stack,
                        visited_nodes,
                        pending_nodes,
                    )
                    if len(stack) > stack_len + 1:  # More than one element was added
                        stack.append("]")
                    else:  # Only one element was added
                        stack.pop(-2)
                for ident in list(pending_nodes):
                    if ident in tmp_pending_nodes:
                        continue
                    pending_nodes.remove(ident)
                    self.walk_nodes(
                        self.find_node(node_id=ident),
                        stack,
                        visited_nodes,
                        pending_nodes,
                    )
            else:
                for ident, link in node.output_links.items():
                    self.walk_nodes(
                        self.find_node(node_id=ident),
                        stack,
                        visited_nodes,
                        pending_nodes,
                    )
            for node in filter(lambda _: _.id >= 0, self.nodes):  # Element nodes
                assert node.get_num_input_links() > 0, node.set_invalid(
                    f"{node.label.strip()} has insufficient input connections!"
                )
                assert node.get_num_output_links() > 0, node.set_invalid(
                    f"{node.label.strip()} has insufficient output connections!"
                )
                assert node.id in visited_nodes, node.set_invalid(
                    f"{node.label.strip()} was not visited!"
                )
            for node in filter(lambda _: _.id < -1, self.nodes):  # Dummy nodes
                num_links_in = node.get_num_input_links()
                num_links_out = node.get_num_output_links()
                assert (num_links_in > 0 and num_links_out > 1) or (
                    num_links_in > 1 and num_links_out > 0
                ), node.set_invalid("A dummy node has insufficient connections!")
                assert node.id in visited_nodes, node.set_invalid(
                    "A dummy node was not visited!"
                )
            assert len(pending_nodes) == 0, "The queue for nodes to visit is not empty!"
            # Make sure all other nodes have been visited
            required_nodes: Set[int] = set(map(lambda _: _.id, self.nodes))
            required_nodes.add(self.cere_node.id)
            assert required_nodes == visited_nodes, "Disconnected node(s) detected!"
            return
        num_links_in = len(node.input_links)
        if node == self.cere_node:  # CE+RE node (i.e. the end point)
            if num_links_in > 1:  # End of a parallel connection
                # Make sure all nodes in the parallel connection have been visited first
                if not set(node.input_links.keys()).issubset(visited_nodes):
                    pending_nodes.add(node.id)
                    return
                elif node.id in pending_nodes:
                    return
                if stack[-1] == "(":
                    stack.pop()
                elif stack[-1] != ")":
                    stack.append(")")
            visited_nodes.add(node.id)
            stack.append("]")
        else:  # Element and dummy nodes
            element: Optional[Element] = node.element
            if num_links_in > 1:  # End of a parallel connection
                # Make sure all nodes in the parallel connection have been visited first
                if not set(node.input_links.keys()).issubset(visited_nodes):
                    pending_nodes.add(node.id)
                    return
                elif node.id in pending_nodes:
                    return
                if stack[-1] != ")":
                    stack.append(")")
            visited_nodes.add(node.id)
            if element is not None:
                assert num_links_out > 0, node.set_invalid(
                    f"{element.get_label()} is missing a connection!"
                )
                stack.append(node.element.to_string(12))  # type: ignore
                # stack.append(node.element.to_string())  # DEBUGGING
            else:
                assert (num_links_in > 0 and num_links_out > 1) or (
                    num_links_in > 1 and num_links_out > 0
                ), node.set_invalid(
                    "Dummy nodes must connect to at least one element on one side "
                    "and at least two elements on the other side!"
                )
            if num_links_out > 1:
                stack.append("(")
                tmp_pending_nodes = set(pending_nodes)
                for ident, link in node.output_links.items():
                    stack.append("[")
                    stack_len = len(stack)
                    self.walk_nodes(
                        self.find_node(node_id=ident),
                        stack,
                        visited_nodes,
                        pending_nodes,
                    )
                    if len(stack) > stack_len + 1:  # More than one element was added
                        stack.append("]")
                    else:  # Only one element was added
                        stack.pop(-2)
                stack.append(")")
                for ident in list(pending_nodes):
                    if ident in tmp_pending_nodes:
                        continue
                    pending_nodes.remove(ident)
                    self.walk_nodes(
                        self.find_node(node_id=ident),
                        stack,
                        visited_nodes,
                        pending_nodes,
                    )
            else:
                for ident, link in node.output_links.items():
                    self.walk_nodes(
                        self.find_node(node_id=ident),
                        stack,
                        visited_nodes,
                        pending_nodes,
                    )

    def generate_circuit(self) -> Tuple[Optional[Circuit], str, List[str]]:
        circuit: Optional[Circuit] = None
        msg: str = "OK"
        stack: List[str] = []
        visited_nodes: Set[int] = set({})
        pending_nodes: Set[int] = set({})
        if self.we_node is not None:
            self.we_node.set_valid()
        if self.cere_node is not None:
            self.cere_node.set_valid()
        list(map(lambda _: _.set_valid(), self.nodes))
        try:
            self.walk_nodes(self.we_node, stack, visited_nodes, pending_nodes)
            circuit = pyimpspec.parse_cdc("".join(stack))
        except AssertionError as e:
            msg = str(e)
        except ParsingError as e:
            msg = str(e)
        return (
            circuit,
            msg,
            stack,
        )

    def generate_nodes(self, input_stack):
        element_stack: List[Union[Connection, Element]] = []
        symbol_stack: List[str] = []
        node_stack: List[Union[Node, List[Node]]] = [self.we_node]

        symbol: str
        element: Union[Connection, Element]
        node: Node
        other: Node
        others: List[Node]
        width: int
        height: int
        w: int
        h: int

        def is_dummy_node(item) -> bool:
            if type(item) is Node:
                return item.id < -1 and item.id > self.we_node.id
            return False

        def pop_input() -> Tuple[str, Union[Element, Series, Parallel]]:
            symbol, element = input_stack.pop(0)
            symbol_stack.append(symbol)
            element_stack.append(element)
            return (
                symbol,
                element,
            )

        def process_series_element(this: Element, x: int, y: int) -> Tuple[int, int]:
            assert isinstance(this, Element), type(element)
            assert type(x) is int
            assert type(y) is int
            nonlocal symbol_stack
            node = self.add_element_node(
                this,
                pos=(
                    x,
                    y,
                ),
            )
            if type(node_stack[-1]) is list:
                if len(element_stack) > 2 and type(element_stack[-3]) is Parallel:
                    if symbol_stack[-2] == ")":
                        for other in node_stack.pop():  # type: ignore
                            self.link_nodes(other, node)
                    else:
                        assert type(node_stack[-2]) is Node
                        if (
                            node_stack[-2] == self.we_node
                            or is_dummy_node(node_stack[-2])
                        ) or (
                            type(node_stack[-2]) is Node
                            and type(node_stack[-1]) is list
                        ):
                            other = node_stack[-2]
                        else:
                            other = node_stack.pop(-2)
                        self.link_nodes(other, node)  # type: ignore
                else:
                    others = node_stack.pop()
                    for other in others:  # type: ignore
                        self.link_nodes(other, node)
            else:
                if node_stack[-1] == self.we_node:
                    other = node_stack[-1]
                else:
                    other = node_stack.pop()
                self.link_nodes(other, node)  # type: ignore
            node_stack.append(node)
            element_stack.pop()
            return (
                1,
                0,
            )

        def process_series_dummy(x: int, y: int) -> Tuple[int, int]:
            assert type(x) is int
            assert type(y) is int
            if (
                element_stack
                and type(element_stack[-1]) is Parallel
                and type(node_stack[-1]) is list
                and symbol_stack[-2] == ")"
                and symbol_stack[-1] == "]"
            ):
                node = self.add_dummy_node(
                    pos=(
                        x,
                        y,
                    )
                )
                for other in node_stack.pop():  # type: ignore
                    self.link_nodes(other, node)
                node_stack.append(node)
                return (
                    1,
                    0,
                )
            return (
                0,
                0,
            )

        def process_series(this: Series, x: int, y: int) -> Tuple[int, int]:
            assert type(this) is Series
            assert type(x) is int
            assert type(y) is int
            nonlocal symbol_stack
            width = 0
            height = 0
            # Main processing of the series connection
            symbol, element = pop_input()
            while element is not this:
                if type(element) is Series:
                    w, h = process_series(element, x + width, y)
                    width += w
                    if h > height:
                        height = h
                elif type(element) is Parallel:
                    w, h = process_parallel(element, x + width, y)
                    width += w
                    if h > height:
                        height = h
                else:
                    w, h = process_series_element(element, x + width, y)
                    width += w
                symbol, element = pop_input()
            assert element_stack.pop() is this
            assert element_stack.pop() is this
            # Add a dummy node, if necessary
            w, h = process_series_dummy(x + width, y)
            width += w
            # Clean up
            if len(node_stack) > 1 and type(node_stack[-2]) is list:
                assert type(node_stack[-1]) is Node
                node_stack[-2].append(node_stack.pop())  # type: ignore
            return (
                max(1, width),
                max(1, height),
            )

        def process_parallel_element(this: Element, x: int, y: int) -> Tuple[int, int]:
            assert isinstance(this, Element), type(element)
            assert type(x) is int
            assert type(y) is int
            nonlocal symbol_stack
            node = self.add_element_node(
                this,
                pos=(
                    x,
                    y,
                ),
            )
            assert type(node_stack[-2]) is Node
            self.link_nodes(node_stack[-2], node)  # type: ignore
            assert type(node_stack[-1]) is list
            node_stack[-1].append(node)  # type: ignore
            element_stack.pop()
            return (
                1,
                1,
            )

        def process_parallel_dummy(x: int, y: int) -> Tuple[int, int]:
            assert type(x) is int
            assert type(y) is int
            if type(node_stack[-1]) is list:
                if len(element_stack) > 2 and type(element_stack[-2]) is Series:
                    node = self.add_dummy_node(
                        pos=(
                            x,
                            y,
                        )
                    )
                    if type(node_stack[-1]) is list:
                        if symbol_stack[-2] == "[" and symbol_stack[-1] == "(":
                            assert type(node_stack[-2]) is Node
                            self.link_nodes(node_stack[-2], node)  # type: ignore
                        else:
                            for other in node_stack.pop():  # type: ignore
                                self.link_nodes(other, node)  # type: ignore
                    else:
                        assert type(node_stack[-2]) is Node
                        self.link_nodes(node_stack[-2], node)  # type: ignore
                    node_stack.append(node)
                elif symbol_stack[-1] == "(" and (
                    symbol_stack[-2] == ")"
                    or (symbol_stack[-2] == "]" and symbol_stack[-3] == ")")
                ):
                    node = self.add_dummy_node(
                        pos=(
                            x,
                            y,
                        )
                    )
                    for other in node_stack.pop():  # type: ignore
                        self.link_nodes(other, node)
                    node_stack.append(node)
                else:
                    return (
                        0,
                        0,
                    )
                return (
                    1,
                    0,
                )
            return (
                0,
                0,
            )

        def process_parallel(this: Parallel, x: int, y: int) -> Tuple[int, int]:
            assert type(this) is Parallel
            assert type(x) is int
            assert type(y) is int
            nonlocal symbol_stack
            width = 0
            height = 0
            # Add a dummy node, if necessary
            dummy_width: int
            dummy_width, _ = process_parallel_dummy(x, y)
            # Main processing of the parallel connection
            node_stack.append([])
            symbol, element = pop_input()
            while element is not this:
                if type(element) is Series:
                    w, h = process_series(element, x + dummy_width, y + height)
                    if w > width:
                        width = w
                    height += h
                elif type(element) is Parallel:
                    w, h = process_parallel(element, x + dummy_width, y + height)
                    if w > width:
                        width = w
                    height += h
                else:
                    w, h = process_parallel_element(
                        element, x + dummy_width, y + height
                    )
                    if w > width:
                        width = w
                    height += h
                symbol, element = pop_input()
            assert element_stack.pop() is this
            assert element_stack.pop() is this
            # Clean up
            if (
                len(node_stack) > 1
                and type(node_stack[-2]) is Node
                and node_stack[-2] != self.we_node
            ):
                node_stack.pop(-2)
            return (
                max(1, width + dummy_width),
                max(1, height),
            )

        input_length: int = len(input_stack)
        symbol, element = pop_input()
        assert symbol == "[" and type(element) is Series
        width, height = process_series(element, 1, 0)
        assert len(element_stack) == 0, element_stack
        assert len(symbol_stack) == input_length
        assert node_stack.pop(0) == self.we_node
        assert len(node_stack) == 1
        if type(node_stack[-1]) is list:
            for node in node_stack[-1]:
                self.link_nodes(node, self.cere_node)
        else:
            self.link_nodes(node_stack[-1], self.cere_node)
        dpg.configure_item(
            self.cere_node.tag,
            pos=(
                self.x_offset + self.x_step * (width + 1),
                self.y_offset,
            ),
        )



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from deareis.data import (
    PlotSeries,
    PlotSettings,
)
from deareis.enums import (
    PlotType,
)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    Dict,
    List,
    Optional,
    Tuple,
)
import dearpygui.dearpygui as dpg
from deareis.gui.plots import Plot
from deareis.gui.data_sets import DataSetsTab
from deareis.gui.drt import DRTTab
from deareis.gui.fitting import FittingTab
from deareis.gui.kramers_kronig import KramersKronigTab
from deareis.gui.overview import OverviewTab
from deareis.gui.plotting import PlottingTab
from deareis.gui.simulation import SimulationTab
from deareis.signals import Signal
import deareis.signals as signals
from deareis.enums import (
    PlotType,
    FitSimOutput,
)
from deareis.data import (
    DRTResult,
    DRTSettings,
    DataSet,
    FitResult,
    FitSettings,
    PlotSettings,
    Project,
    SimulationResult,
    SimulationSettings,
    TestResult,
    TestSettings,
)
import deareis.themes as themes
from deareis.keybindings import Context


class ProjectTab:
    def __init__(self, uuid: str, tab_bar: int, state):
        assert type(uuid) is str, uuid
        assert type(tab_bar) is int, tab_bar
        self.uuid: str = uuid
        self.tab: int = dpg.generate_uuid()
        with dpg.tab(
            label="Project",
            order_mode=dpg.mvTabOrder_Reorderable,
            user_data=uuid,
            tag=self.tab,
            parent=tab_bar,
        ):
            tab_lookup: dict = {}
            self.context_lookup: Dict[int, Context] = {}

            def select_tab(tab_id: int):
                tab = tab_lookup.get(tab_id)
                assert tab is not None
                tab.resize(
                    width=dpg.get_viewport_width(),
                    height=dpg.get_viewport_height(),
                )
                if hasattr(tab, "queued_update") and tab.queued_update is not None:
                    tab.queued_update()

            self.tab_bar: int = dpg.generate_uuid()
            with dpg.tab_bar(
                callback=lambda s, a, u: select_tab(a),
                tag=self.tab_bar,
            ):
                self.overview_tab: OverviewTab = OverviewTab()
                self.data_sets_tab: DataSetsTab = DataSetsTab()
                self.kramers_kronig_tab: KramersKronigTab = KramersKronigTab(state)
                self.drt_tab: DRTTab = DRTTab(state)
                self.fitting_tab: FittingTab = FittingTab(state)
                self.simulation_tab: SimulationTab = SimulationTab(state)
                self.plotting_tab: PlottingTab = PlottingTab(state)
            tab_lookup.update(
                {
                    self.overview_tab.tab: self.overview_tab,
                    self.data_sets_tab.tab: self.data_sets_tab,
                    self.kramers_kronig_tab.tab: self.kramers_kronig_tab,
                    self.drt_tab.tab: self.drt_tab,
                    self.fitting_tab.tab: self.fitting_tab,
                    self.simulation_tab.tab: self.simulation_tab,
                    self.plotting_tab.tab: self.plotting_tab,
                }
            )
            self.context_lookup.update(
                {
                    self.overview_tab.tab: Context.OVERVIEW_TAB,
                    self.data_sets_tab.tab: Context.DATA_SETS_TAB,
                    self.kramers_kronig_tab.tab: Context.KRAMERS_KRONIG_TAB,
                    self.drt_tab.tab: Context.DRT_TAB,
                    self.fitting_tab.tab: Context.FITTING_TAB,
                    self.simulation_tab.tab: Context.SIMULATION_TAB,
                    self.plotting_tab.tab: Context.PLOTTING_TAB,
                }
            )

            def visibility_handler(tab):
                if not dpg.does_item_exist(tab.tab):
                    return
                elif not tab.is_visible():
                    return
                elif hasattr(tab, "queued_update") and tab.queued_update is not None:
                    tab.queued_update()

            for tab in tab_lookup.values():
                if not (hasattr(tab, "is_visible") and hasattr(tab, "visibility_item")):
                    continue
                item_handler: int = dpg.generate_uuid()
                with dpg.item_handler_registry(tag=item_handler):
                    dpg.add_item_visible_handler(
                        callback=lambda s, a, u: visibility_handler(u),
                        user_data=tab,
                    )
                dpg.bind_item_handler_registry(tab.visibility_item, item_handler)

    def resize(self, width: int, height: int):
        assert type(width) is int and width > 0, width
        assert type(height) is int and height > 0, height
        self.data_sets_tab.resize(width, height)
        self.kramers_kronig_tab.resize(width, height)
        self.drt_tab.resize(width, height)
        self.fitting_tab.resize(width, height)
        self.simulation_tab.resize(width, height)
        self.plotting_tab.resize(width, height)

    def select_next_tab(self):
        current_tab: int = dpg.get_value(self.tab_bar)
        tabs: List[int] = dpg.get_item_children(self.tab_bar, slot=1)
        while tabs:
            tab: int = tabs.pop(0)
            if tab == current_tab:
                break
        if not tabs:
            tabs = dpg.get_item_children(self.tab_bar, slot=1)
        tab = tabs[0]
        dpg.set_value(self.tab_bar, tab)

    def select_previous_tab(self):
        current_tab: int = dpg.get_value(self.tab_bar)
        tabs: List[int] = dpg.get_item_children(self.tab_bar, slot=1)
        while tabs:
            tab: int = tabs.pop()
            if tab == current_tab:
                break
        if not tabs:
            tabs = dpg.get_item_children(self.tab_bar, slot=1)
        tab = tabs[-1]
        dpg.set_value(self.tab_bar, tab)

    def get_active_context(self) -> Context:
        return self.context_lookup[dpg.get_value(self.tab_bar)]

    def get_active_data_set(
        self, context: Optional[Context] = None
    ) -> Optional[DataSet]:
        assert type(context) is Context or context is None, context
        tag: Optional[int] = None
        if context is not None:
            tag = {
                Context.DATA_SETS_TAB: self.data_sets_tab.delete_button,
                Context.KRAMERS_KRONIG_TAB: self.kramers_kronig_tab.perform_test_button,
                Context.DRT_TAB: self.drt_tab.perform_drt_button,
                Context.FITTING_TAB: self.fitting_tab.perform_fit_button,
                Context.SIMULATION_TAB: self.simulation_tab.perform_sim_button,
            }.get(context)
        else:
            tag = {
                self.data_sets_tab.tab: self.data_sets_tab.delete_button,
                self.kramers_kronig_tab.tab: self.kramers_kronig_tab.perform_test_button,
                self.drt_tab.tab: self.drt_tab.perform_drt_button,
                self.fitting_tab.tab: self.fitting_tab.perform_fit_button,
                self.simulation_tab.tab: self.simulation_tab.perform_sim_button,
            }.get(dpg.get_value(self.tab_bar))
        if tag is None:
            return None
        return dpg.get_item_user_data(tag)

    def get_active_test(self) -> Optional[TestResult]:
        return dpg.get_item_user_data(self.kramers_kronig_tab.delete_button).get("test")

    def get_active_drt(self) -> Optional[DRTResult]:
        return dpg.get_item_user_data(self.drt_tab.delete_button).get("drt")

    def get_active_fit(self) -> Optional[FitResult]:
        return dpg.get_item_user_data(self.fitting_tab.delete_button).get("fit")

    def get_active_simulation(self) -> Optional[SimulationResult]:
        return dpg.get_item_user_data(self.simulation_tab.delete_button)

    def get_active_plot(self) -> Optional[PlotSettings]:
        return self.plotting_tab.get_active_plot()

    def select_overview_tab(self):
        dpg.set_value(self.tab_bar, self.overview_tab.tab)

    def select_data_sets_tab(self):
        dpg.set_value(self.tab_bar, self.data_sets_tab.tab)

    def select_kramers_kronig_tab(self):
        dpg.set_value(self.tab_bar, self.kramers_kronig_tab.tab)

    def select_drt_tab(self):
        dpg.set_value(self.tab_bar, self.drt_tab.tab)

    def select_fitting_tab(self):
        dpg.set_value(self.tab_bar, self.fitting_tab.tab)

    def select_simulation_tab(self):
        dpg.set_value(self.tab_bar, self.simulation_tab.tab)

    def select_plotting_tab(self):
        dpg.set_value(self.tab_bar, self.plotting_tab.tab)

    def set_label(self, label: str):
        assert type(label) is str and label != "", label
        dpg.set_item_label(self.tab, label)
        self.overview_tab.set_label(label)

    def get_notes(self) -> str:
        return self.overview_tab.get_notes()

    def set_notes(self, notes: str):
        assert type(notes) is str, notes
        self.overview_tab.set_notes(notes)

    def set_dirty(self, state: bool):
        dpg.bind_item_theme(self.tab, themes.tab.dirty if state else themes.tab.clean)

    def populate_data_sets(self, project: Project):
        assert type(project) is Project, project
        data_sets: List[DataSet] = project.get_data_sets()
        lookup: Dict[str, DataSet] = {_.get_label(): _ for _ in data_sets}
        labels: List[str] = list(lookup.keys())
        self.data_sets_tab.populate_data_sets(labels, lookup)
        self.kramers_kronig_tab.populate_data_sets(labels, lookup)
        self.drt_tab.populate_data_sets(labels, lookup)
        self.fitting_tab.populate_data_sets(labels, lookup)
        self.simulation_tab.populate_data_sets(labels, lookup)

    def populate_tests(self, project: Project, data: Optional[DataSet]):
        assert type(project) is Project, project
        assert type(data) is DataSet or data is None, data
        self.kramers_kronig_tab.populate_tests(
            {_.get_label(): _ for _ in project.get_tests(data)}
            if data is not None
            else {},
            data,
        )

    def populate_drts(self, project: Project, data: Optional[DataSet]):
        assert type(project) is Project, project
        assert type(data) is DataSet or data is None, data
        self.drt_tab.populate_drts(
            {_.get_label(): _ for _ in project.get_drts(data)}
            if data is not None
            else {},
            data,
        )

    def populate_fits(self, project: Project, data: Optional[DataSet]):
        assert type(project) is Project, project
        assert type(data) is DataSet or data is None, data
        lookup: Dict[str, FitResult] = (
            {_.get_label(): _ for _ in project.get_fits(data)}
            if data is not None
            else {}
        )
        self.fitting_tab.populate_fits(lookup, data)
        # NOTE: 'lookup' is modified in-place by the fitting tab!
        # The labels are formatted so that the timestamp starts at the same point
        # in the various labels, which should improve readability as it concerns
        # the CDC of each result.
        self.drt_tab.populate_fits(lookup)

    def populate_simulations(self, project: Project):
        assert type(project) is Project, project
        self.simulation_tab.populate_simulations(
            {_.get_label(): _ for _ in project.get_simulations()}
        )

    def populate_plots(self, project: Project):
        assert type(project) is Project, project
        self.plotting_tab.populate_plots(
            {_.get_label(): _ for _ in project.get_plots()}
        )

    def select_data_set(self, data: Optional[DataSet]):
        assert type(data) is DataSet or data is None, data
        self.data_sets_tab.select_data_set(data)
        self.kramers_kronig_tab.select_test_result(None, data)
        self.drt_tab.select_drt_result(None, data)
        self.fitting_tab.select_fit_result(None, data)

    def get_next_data_set(self, context: Context) -> Optional[DataSet]:
        if context == Context.DATA_SETS_TAB:
            return self.data_sets_tab.get_next_data_set()
        elif context == Context.KRAMERS_KRONIG_TAB:
            return self.kramers_kronig_tab.get_next_data_set()
        elif context == Context.DRT_TAB:
            return self.drt_tab.get_next_data_set()
        elif context == Context.FITTING_TAB:
            return self.fitting_tab.get_next_data_set()
        return None

    def get_previous_data_set(self, context: Context) -> Optional[DataSet]:
        if context == Context.DATA_SETS_TAB:
            return self.data_sets_tab.get_previous_data_set()
        elif context == Context.KRAMERS_KRONIG_TAB:
            return self.kramers_kronig_tab.get_previous_data_set()
        elif context == Context.DRT_TAB:
            return self.drt_tab.get_previous_data_set()
        elif context == Context.FITTING_TAB:
            return self.fitting_tab.get_previous_data_set()
        return None

    def get_next_simulation_data_set(self) -> Optional[DataSet]:
        return self.simulation_tab.get_next_data_set()

    def get_previous_simulation_data_set(self) -> Optional[DataSet]:
        return self.simulation_tab.get_previous_data_set()

    def get_next_test_result(self) -> Optional[TestResult]:
        return self.kramers_kronig_tab.get_next_result()

    def get_previous_test_result(self) -> Optional[TestResult]:
        return self.kramers_kronig_tab.get_previous_result()

    def get_next_drt_result(self) -> Optional[DRTResult]:
        return self.drt_tab.get_next_result()

    def get_previous_drt_result(self) -> Optional[DRTResult]:
        return self.drt_tab.get_previous_result()

    def get_next_fit_result(self) -> Optional[FitResult]:
        return self.fitting_tab.get_next_result()

    def get_previous_fit_result(self) -> Optional[FitResult]:
        return self.fitting_tab.get_previous_result()

    def get_next_simulation_result(self) -> Optional[SimulationResult]:
        return self.simulation_tab.get_next_result()

    def get_previous_simulation_result(self) -> Optional[SimulationResult]:
        return self.simulation_tab.get_previous_result()

    def get_next_plot(self) -> Optional[PlotSettings]:
        return self.plotting_tab.get_next_plot()

    def get_previous_plot(self) -> Optional[PlotSettings]:
        return self.plotting_tab.get_previous_plot()

    def get_next_plot_type(self) -> Optional[PlotType]:
        return self.plotting_tab.get_next_plot_type()

    def get_previous_plot_type(self) -> Optional[PlotType]:
        return self.plotting_tab.get_previous_plot_type()

    def select_test_result(self, test: TestResult, data: DataSet):
        assert type(test) is TestResult, test
        assert type(data) is DataSet, data
        self.kramers_kronig_tab.select_test_result(test, data)

    def select_drt_result(self, drt: DRTResult, data: DataSet):
        assert type(drt) is DRTResult, drt
        assert type(data) is DataSet, data
        self.drt_tab.select_drt_result(drt, data)

    def select_fit_result(self, fit: FitResult, data: DataSet):
        assert type(fit) is FitResult, fit
        assert type(data) is DataSet, data
        self.fitting_tab.select_fit_result(fit, data)

    def select_simulation_result(
        self, simulation: Optional[SimulationResult], data: Optional[DataSet]
    ):
        assert type(simulation) is SimulationResult or simulation is None, simulation
        assert type(data) is DataSet or data is None, data
        self.simulation_tab.select_simulation_result(simulation, data)

    def select_plot(
        self,
        settings: PlotSettings,
        data_sets: List[DataSet],
        tests: Dict[str, List[TestResult]],
        drts: Dict[str, List[DRTResult]],
        fits: Dict[str, List[FitResult]],
        simulations: List[SimulationResult],
        adjust_limits: bool,
        plot_only: bool,
    ):
        assert type(settings) is PlotSettings, settings
        assert type(data_sets) is list, data_sets
        assert type(tests) is dict, tests
        assert type(drts) is dict, drts
        assert type(fits) is dict, fits
        assert type(simulations) is list, simulations
        assert type(adjust_limits) is bool, adjust_limits
        assert type(plot_only) is bool, plot_only
        self.plotting_tab.select_plot(
            settings,
            data_sets,
            tests,
            drts,
            fits,
            simulations,
            adjust_limits,
            plot_only,
        )

    def select_plot_type(self, plot_type: PlotType):
        assert type(plot_type) is PlotType, plot_type
        self.plotting_tab.select_plot_type(plot_type)

    def set_series_label(self, uuid: str, label: Optional[str]):
        assert type(uuid) is str, uuid
        assert type(label) is str or label is None, label
        self.plotting_tab.set_series_label(uuid, label)

    def get_test_settings(self) -> TestSettings:
        return self.kramers_kronig_tab.get_settings()

    def set_test_settings(self, settings: TestSettings):
        assert type(settings) is TestSettings, settings
        self.kramers_kronig_tab.set_settings(settings)

    def get_fit_settings(self) -> FitSettings:
        return self.fitting_tab.get_settings()

    def set_fit_settings(self, settings: FitSettings):
        assert type(settings) is FitSettings, settings
        self.fitting_tab.set_settings(settings)

    def get_drt_settings(self) -> DRTSettings:
        return self.drt_tab.get_settings()

    def set_drt_settings(self, settings: DRTSettings):
        assert type(settings) is DRTSettings, settings
        self.drt_tab.set_settings(settings)

    def get_simulation_settings(self) -> SimulationSettings:
        return self.simulation_tab.get_settings()

    def set_simulation_settings(self, settings: SimulationSettings):
        assert type(settings) is SimulationSettings, settings
        self.simulation_tab.set_settings(settings)

    def show_enlarged_nyquist(self):
        {
            self.data_sets_tab.tab: self.data_sets_tab.show_enlarged_nyquist,
            self.kramers_kronig_tab.tab: self.kramers_kronig_tab.show_enlarged_nyquist,
            self.fitting_tab.tab: self.fitting_tab.show_enlarged_nyquist,
            self.simulation_tab.tab: self.simulation_tab.show_enlarged_nyquist,
        }.get(dpg.get_value(self.tab_bar))()

    def show_enlarged_bode(self):
        {
            self.data_sets_tab.tab: self.data_sets_tab.show_enlarged_bode,
            self.kramers_kronig_tab.tab: self.kramers_kronig_tab.show_enlarged_bode,
            self.fitting_tab.tab: self.fitting_tab.show_enlarged_bode,
            self.simulation_tab.tab: self.simulation_tab.show_enlarged_bode,
        }.get(dpg.get_value(self.tab_bar))()

    def show_enlarged_residuals(self):
        {
            self.kramers_kronig_tab.tab: self.kramers_kronig_tab.show_enlarged_residuals,
            self.drt_tab.tab: self.drt_tab.show_enlarged_residuals,
            self.fitting_tab.tab: self.fitting_tab.show_enlarged_residuals,
        }.get(dpg.get_value(self.tab_bar))()

    def show_enlarged_drt(self):
        {
            self.drt_tab.tab: self.drt_tab.show_enlarged_drt,
        }.get(dpg.get_value(self.tab_bar))()

    def show_enlarged_impedance(self):
        {
            self.drt_tab.tab: self.drt_tab.show_enlarged_impedance,
        }.get(dpg.get_value(self.tab_bar))()

    def update_plots(
        self,
        settings: PlotSettings,
        data_sets: List[DataSet],
        tests: Dict[str, List[TestResult]],
        drts: Dict[str, List[DRTResult]],
        fits: Dict[str, List[FitResult]],
        simulations: List[SimulationResult],
    ):
        assert type(settings) is PlotSettings, settings
        assert type(data_sets) is list, data_sets
        assert type(tests) is dict, tests
        assert type(drts) is dict, drts
        assert type(fits) is dict, fits
        assert type(simulations) is list, simulations
        self.plotting_tab.plot_series(
            data_sets,
            tests,
            drts,
            fits,
            simulations,
            settings,
            adjust_limits=False,
        )

    def get_drt_plot(self, context: Context) -> Optional[Plot]:
        if context == Context.DRT_TAB:
            return self.drt_tab.drt_plot
        return None

    def get_impedance_plot(self, context: Context) -> Optional[Plot]:
        if context == Context.DRT_TAB:
            return self.drt_tab.impedance_plot
        return None

    def get_nyquist_plot(self, context: Context) -> Optional[Plot]:
        if context == Context.KRAMERS_KRONIG_TAB:
            return self.kramers_kronig_tab.nyquist_plot
        elif context == Context.FITTING_TAB:
            return self.fitting_tab.nyquist_plot
        elif context == Context.SIMULATION_TAB:
            return self.simulation_tab.nyquist_plot
        return None

    def get_bode_plot(self, context: Context) -> Optional[Plot]:
        if context == Context.KRAMERS_KRONIG_TAB:
            return self.kramers_kronig_tab.bode_plot_horizontal
        elif context == Context.FITTING_TAB:
            return self.fitting_tab.bode_plot_horizontal
        elif context == Context.SIMULATION_TAB:
            return self.simulation_tab.bode_plot_horizontal
        return None

    def get_residuals_plot(self, context: Context) -> Optional[Plot]:
        if context == Context.KRAMERS_KRONIG_TAB:
            return self.kramers_kronig_tab.residuals_plot
        elif context == Context.DRT_TAB:
            return self.drt_tab.residuals_plot
        elif context == Context.FITTING_TAB:
            return self.fitting_tab.residuals_plot
        return None

    def get_active_output(self, context: Context) -> Optional[FitSimOutput]:
        if context == Context.FITTING_TAB:
            return self.fitting_tab.get_active_output()
        elif context == Context.SIMULATION_TAB:
            return self.simulation_tab.get_active_output()
        return None

    def get_filtered_plot_series(
        self,
    ) -> Tuple[
        List[DataSet],
        List[TestResult],
        List[DRTResult],
        List[FitResult],
        List[SimulationResult],
    ]:
        string: str = self.plotting_tab.get_filter_string()
        return (
            self.plotting_tab.possible_data_sets.filter(string, False),
            self.plotting_tab.possible_tests.filter(string, False),
            self.plotting_tab.possible_drts.filter(string, False),
            self.plotting_tab.possible_fits.filter(string, False),
            self.plotting_tab.possible_simulations.filter(string, False),
        )

    def has_active_input(self, context: Optional[Context] = None) -> bool:
        assert type(context) is Context or context is None, context
        if context is None:
            context = self.get_active_context()
        if context == Context.OVERVIEW_TAB:
            return self.overview_tab.has_active_input()
        elif context == Context.DATA_SETS_TAB:
            return self.data_sets_tab.has_active_input()
        elif context == Context.KRAMERS_KRONIG_TAB:
            return self.kramers_kronig_tab.has_active_input()
        elif context == Context.DRT_TAB:
            return self.drt_tab.has_active_input()
        elif context == Context.FITTING_TAB:
            return self.fitting_tab.has_active_input()
        elif context == Context.SIMULATION_TAB:
            return self.simulation_tab.has_active_input()
        elif context == Context.PLOTTING_TAB:
            return self.plotting_tab.has_active_input()
        return False
